/**
 * Example formatting from fast-components-react-base
 */
import Foo from "./foo";
export { Foo };
export * from "./foo";

/**
 * Example formatting from fast-component-react-msft
 */
export * from "./bar";
